unit cstring;
{
    C/C++ string.h and C++ string object
    (C) Copyright 1996  John Webster Small
    All rights reserved
}

interface

uses sysutils, DB, DBTables, StdCtrls, classes;

const

    CSTR_MAX = 65520;
    CSTR_NOTFOUND = CSTR_MAX;
    CSTR_WS = ' '#9#10#12#13;
    CSTR_AVE_LN_LEN = 30;

type

    ECstrException = class(Exception);

    CSTR_format = ( Left, Center, Right, Both );

    cstr = class(TPersistent)
      public
        constructor Create;
        constructor Create_string(const s : string);
        constructor Create_extra(extra : cardinal);
        constructor Create_cstr(s : cstr);
        constructor Create_cstr_extra
            (s : cstr; extra : cardinal);
        constructor Create_PChar(S : PChar);
        constructor Create_PChar_extra
            (S : PChar; extra : cardinal);
        constructor Create_char_fill
            (c : char; nfill : cardinal);
        constructor Create_char_fill_extra
            (c : char; nfill, extra : cardinal);
        constructor Create_Strm(var inStrm : TextFile);
        constructor Create_File(const filename : string);
        destructor  Destroy; override;
        function wordWrap(atColumn : cardinal) : cstr;
        function str : PChar;
        function StrPas : string;
      private
        procedure setAsString(const value : string);
        function  getChar(Index: cardinal) : char;
        procedure putChar(Index: cardinal; c : char);
      public
        property AsString : string read StrPas write setAsString;
        property CharArray[Index : cardinal] : char
          read getChar write putChar; default;
        function cpy(S : PChar) : cstr;
        function cpy_extra(S : PChar; extra : cardinal) : cstr;
        function cpy_string(const s : string) : cstr;
        function ncpy(S : PChar; n : cardinal) : cstr;
        function ncpy_extra
            (S : PChar; n, extra : cardinal) : cstr;
        function ncpy_char_fill
            (c : char; nfill : cardinal) : cstr;
        function ncpy_char_fill_extra
            (c : char; nfill, extra : cardinal) : cstr;
        function ncpy_fmt(n : cardinal; format : PChar;
            const Args : array of const) : cstr;
        function cat(S : PChar) : cstr;
        function cat_extra(S : PChar; extra : cardinal) : cstr;
        function cat_string(const s : string) : cstr;
        function ncat(S : PChar; n : cardinal) : cstr;
        function ncat_extra
            (S : PChar; n, extra : cardinal) : cstr;
        function ncat_char_fill
            (c : char; nfill : cardinal) : cstr;
        function ncat_char_fill_extra
            (c : char; nfill, extra : cardinal) : cstr;
        function ncat_fmt(n : cardinal; format : PChar;
            const Args : array of const) : cstr;
        function cat_itoa
          (value, radix : integer) : cstr;
        function cat_ltoa
          (value : longint; radix : integer) : cstr;
        function cat_ultoa
          (value : cardinal; radix : integer) : cstr;
        function atof : extended;
        function strtod(suffix : cstr) : extended;
        function strtol
          (base : integer; suffix : cstr) : longint;
        function strtoul
          (base : integer; suffix : cstr) : cardinal;
        function nins
            (pos : cardinal; S : PChar; n : cardinal) : cstr;
        function nins_extra
            (pos : cardinal; S : PChar; n, extra : cardinal) : cstr;
        function nins_char_fill
            (pos : cardinal; c : char; nfill : cardinal) : cstr;
        function nins_char_fill_extra
            (pos : cardinal; c : char;
             nfill, extra : cardinal) : cstr;
        function ins(pos : cardinal; S : PChar) : cstr;
        function ins_extra
            (pos : cardinal; S : PChar; extra : cardinal) : cstr;
        function prefix(S : PChar) : cstr;
        function prefix_extra
            (S : PChar; extra : cardinal) : cstr;
        function prefix_string(const s : string) : cstr;
        function prefix_char_fill
            (c : char; nfill : cardinal) : cstr;
        function prefix_char_fill_extra
            (c : char; nfill, extra : cardinal) : cstr;
        function nprefix_fmt(n : cardinal; format : PChar;
            const Args : array of const) : cstr;
        function ncpy2
            (pos : cardinal; S : PChar; n : cardinal) : cstr;
        function ncpy2_extra
            (pos : cardinal; S : PChar; n, extra : cardinal) : cstr;
        function ncpy2_char_fill
            (pos : cardinal; c : char; nfill : cardinal) : cstr;
        function ncpy2_char_fill_extra
            (pos : cardinal; c : char;
             nfill, extra : cardinal) : cstr;
        function cpy2(pos : cardinal; S : PChar) : cstr;
        function cpy2_extra
            (pos : cardinal; S : PChar; extra : cardinal) : cstr;
        function rmv(pos, n : cardinal) : cstr;
        function rmvTail(pos : cardinal) : cstr;
        function substr
            (dst : cstr; pos, n : cardinal) : cstr;
        function subString(pos, n : cardinal) : string;
        function tailString(pos : cardinal) : string;
        function substr_extra
            (dst : cstr; pos, n : cardinal;
             extra : cardinal) : cstr;
        function extract
            (dst : cstr; pos, n : cardinal) : cstr;
        function extract_extra
            (dst : cstr; pos, n : cardinal;
             extra : cardinal) : cstr;
        function replace_PChar_fill
            (pos, n : cardinal; S : PChar) : cstr;
        function replace_PChar_fill_extra
            (pos, n : cardinal; S : PChar; extra : cardinal) : cstr;
        function replace_char_fill
            (pos, n : cardinal; c : char) : cstr;
        function replace_char_fill_extra
            (pos, n : cardinal; c : char; extra : cardinal) : cstr;
        function replace(P, S : PChar) : cstr;
        function replace_extra
            (P, S : PChar; extra : cardinal) : cstr;
        function ireplace(P, S : PChar) : cstr;
        function ireplace_extra
            (P, S : PChar; extra : cardinal) : cstr;
        function replaceAll(P, S : PChar) : cstr;
        function replaceAll_extra
            (P, S : PChar; extra : cardinal) : cstr;
        function removeAllEOL : cstr;  { use after assign from TStrings }
        function ireplaceAll(P, S : PChar) : cstr;
        function ireplaceAll_extra
            (P, S : PChar; extra : cardinal) : cstr;
        function trim(f : CSTR_format; strip : PChar) : cstr;
        function trimBothWS : cstr;
        function pad(fieldWidth : cardinal;
            f : CSTR_format; fillch : char) : cstr;
        function padLeftSpaces(fieldWidth : cardinal) : cstr;
        function justify(fieldWidth : cardinal;
            f : CSTR_format; fillch : char) : cstr;
        function justifyLeftSpaces(fieldWidth : cardinal) : cstr;
        function idx(S : PChar) : cardinal;
        function idx_chr(c : char) : cardinal;
        function chr(c : char) : PChar;
        function idx_rchr(c : char) : cardinal;
        function rchr(c : char) : PChar;
        function spn(S : PChar) : cardinal;
        function cspn(S : PChar) : cardinal;
        function idx_brk(S : PChar) : cardinal;
        function pbrk(S : PChar) : PChar;
        function idx_str(S : PChar) : cardinal;
        function strstr(S : PChar) : PChar;
        function idx_str_start
            (S : PChar; start : cardinal) : cardinal;
        function idx_stri(S : PChar) : cardinal;
        function strstri(S : PChar) : PChar;
        function idx_stri_start
            (S : PChar; start : cardinal) : cardinal;
        function tolower : cstr;
        function toupper : cstr;
        function toproper : cstr;
        function len : cardinal;
        function lenReCalc : cstr;
        function vacancy : cardinal;
        function limit : cardinal;
        function setLimit(limit_ : cardinal) : cstr;
        function atLeastLimit(minlimit : cardinal) : cstr;
        function pack : cstr;
        function unpack : cstr;
        function trunc(len_ : cardinal) : cstr;
        function clr : cstr;
        procedure clear;
        function cmp(S : PChar) : integer;
        function icmp(S : PChar) : integer;
        function ncmp(S : PChar; n : cardinal) : integer;
        function nicmp(S : PChar; n : cardinal) : integer;
        function printOnFile(var outStrm : TextFile) : cstr;
        function storeOnFile(var outStrm : TextFile) : cstr;
        function storeInString(var s : string) : cstr;
        function storeInStrings(strs : TStrings) : cstr;
        function storeOnStream(strm : TStream) : cstr;
        function storeInMemo(Memo : TCustomMemo) : cstr;
        function storeInBlobField(Field : TBlobField) : cstr;
          {
            DataSet.Insert;
            DataSet.ClearFields;
            field.storeInDataSet(DataSet,ftMemo,'MyFieldName');
            ...
            DataSet.Post;
          }
        function storeInDataSet(DataSet : TDataSet;
          FieldType : TFieldType; const FieldName : string) : cstr;
        function readFromFile(var inStrm : TextFile) : cstr;
        function readFromString(str : string) : cstr;
        function readFromStrings(strs : TStrings) : cstr;
        function readFromStream(strm : TStream) : cstr;
        function readFromMemo(Memo : TCustomMemo) : cstr;
        function readFromBlobField(Field : TBlobField) : cstr;
        function readFromDataSet(DataSet : TDataSet;
          FieldType : TFieldType; const FieldName : string) : cstr;
        function SaveToFile(const FileName : string) : cstr;
        function SaveToStream(Stream : TStream) : cstr;
        function LoadFromFile(const FileName : string) : cstr;
        function LoadFromStream(Stream : TStream) : cstr;
        function ignore(var inStrm : TextFile) : cstr;
        function getline(var inStrm : TextFile) : cstr;
        function readBuf(var Buffer; Count, pos : cardinal): cardinal;
        function writeBuf(const Buffer; Count, pos : cardinal): cardinal;
        procedure Assign(Source : TPersistent);
      protected
        procedure AssignTo(Dest : TPersistent); override;
        procedure DefineProperties(Filer : TFiler); override;
      private
        _S : PChar;
        _len, _limit : cardinal;
        procedure _ReadStr(Stream : TStream);
        procedure _WriteStr(Stream : TStream);
        procedure _init;
        procedure _done;
        function _ncpy(S : PChar; n, extra : cardinal) : cstr;
        function _ncat(S : PChar; n, extra : cardinal) : cstr;
        procedure _init_fill(c : char; nfill, extra : cardinal);
        function _fill(c : char; nfill, extra : cardinal) : cstr;
        class function fillChar(c : char; nfill: cardinal) : PChar;
    end;

function strcpy(s, ct : PChar) : PChar;
function strncpy(s, ct : PChar; n : cardinal) : PChar;
{ function strcat(s, ct : PChar) : PChar; }
function strncat(s, ct : PChar; n : cardinal) : PChar;
function strcmp(cs, ct : PChar) : integer;
function stricmp(cs, ct : PChar) : integer;
function strncmp(cs, ct : PChar; n : cardinal) : integer;
function strncmpi(cs, ct : PChar; n : cardinal) : integer;
function strchr(cs : PChar; c : char) : PChar;
function strrchr(cs : PChar; c: char) : PChar;
function strspn(cs, ct : PChar) : cardinal;
function strcspn(cs, ct : PChar) : cardinal;
function strpbrk(cs, ct : PChar) : PChar;
function strstr(cs, ct : PChar) : PChar;
function strlen(cs : PChar) : cardinal;
function strtok(input, delimiters : PChar) : PChar;

function memcpy(s, ct : pointer; n : cardinal) : pointer;
function memmove(s, ct : pointer; n : cardinal) : pointer;
function memcmp(cs, ct : pointer; n : cardinal) : integer;
function memchr(cs : pointer; c : char; n : cardinal)
  : pointer;
function memset(s : pointer; c : char; n : cardinal)
  : pointer;

function itoa(value, radix : integer) : string;
function ltoa(value : longint; radix : integer) : string;
function ultoa(value : longint; radix : integer) : string;
function atof(S : PChar) : extended;
function atoi(S : PChar) : integer;
function atol(S : PChar) : longint;
function strtod(S : PChar; var endp : PChar) : extended;
function strtol(S : PChar; var endp : PChar;
  base : integer) : longint;
function strtoul(S : PChar; var endp : PChar;
  base : integer) : cardinal;

{ Not recommended! }
function StrPChar(const s: string) : PChar;

procedure AssignFile_cstr
  (var F : TextFile; buf: cstr);

type

    CstrStream = class (TStream)
    private
      fCstr : cstr;
      pos : cardinal;
    public
      destructor Destroy; override;
      function Read(var Buffer; Count: longint): longint; override;
      function Write(const Buffer; Count: longint): longint; override;
      function Seek(Offset: longint; Origin: word): longint; override;
      procedure LoadFromStream(Stream: TStream);
      procedure LoadFromFile(const FileName: string);
      procedure SaveToStream(Stream: TStream);
      procedure SaveToFile(const FileName: string);
      procedure Clear;
      property pCstr : cstr read fCstr;
    end;

{ Pascal String routines }

procedure pstrcpy(var s : string; const ct : string);
procedure pstrncpy(var s : string; const ct : string; n : integer);
procedure pstrcat(var s : string; const ct : string);
procedure pstrncat(var s : string; const ct : string; n : integer);
function  pstrcmp(const cs, ct : string) : integer;
function  pstricmp(const cs, ct : string) : integer;
function  pstrncmp(const cs, ct : string; n : integer) : integer;
function  pstrncmpi(const cs, ct : string; n : integer) : integer;
function  pstrchr(const cs : string; c : char) : integer;
function  pstrrchr(const cs : string; c: char) : integer;
function  pstrspn(const cs, ct : string) : integer;
function  pstrcspn(const cs, ct : string) : integer;
function  pstrbrk(const cs, ct : string) : integer;
function  pstrstr(const cs, ct : string) : integer;
function  pstrlen(const cs : string) : integer;
procedure ptrimBothWS(var s : string);

function patof(const s : string) : extended;
function patoi(const s : string) : integer;
function patol(const s : string) : longint;
function pstrtod(const s : string; var ends : string) : extended;
function pstrtol(const s : string; var ends : string;
  base : integer) : longint;
function pstrtoul(const s : string; var ends : string;
  base : integer) : cardinal;

implementation

uses ctype;

var

    strtok_input, strtok_found, fill_char, PCharBuf : cstr;

constructor cstr.Create;
begin
    inherited Create;
    _init;
end;

constructor cstr.Create_extra(extra : cardinal);
begin
    inherited Create;
    _init;
    _ncat(nil,0,extra);
end;

constructor cstr.Create_string(const s : string);
begin
    inherited Create;
    _init;
    _ncpy(PChar(@s[1]),cardinal(Length(s)),0);
end;

constructor cstr.Create_cstr(s : cstr);
begin
    inherited Create;
    _init;
    _ncat(s.str,CSTR_MAX,0);
end;

constructor cstr.Create_cstr_extra
    (s : cstr; extra : cardinal);
begin
    inherited Create;
    _init;
    _ncat(s.str,CSTR_MAX,extra);
end;

constructor cstr.Create_PChar(S : PChar);
begin
    inherited Create;
    _init;
    _ncat(S,CSTR_MAX,0);
end;

constructor cstr.Create_PChar_extra
    (S : PChar; extra : cardinal);
begin
    inherited Create;
    _init;
    _ncat(S,CSTR_MAX,extra);
end;

constructor cstr.Create_char_fill
    (c : char; nfill : cardinal);
begin
    inherited Create;
    _init_fill(c,nfill,0);
end;

constructor cstr.Create_char_fill_extra
    (c : char; nfill, extra : cardinal);
begin
    inherited Create;
    _init_fill(c,nfill,extra);
end;

constructor cstr.Create_Strm(var inStrm : TextFile);
begin
    inherited Create;
    _init;
    readFromFile(inStrm);
end;

constructor cstr.Create_File(const filename : string);
begin
   inherited Create;
   _init;
   LoadFromFile(filename);
end;

destructor cstr.Destroy;
begin
    _done;
    inherited Destroy;
end;

function cstr.wordWrap(atColumn : cardinal) : cstr;
var i, nl, nlCol : cardinal;
begin
  i := 0;
  while i < _len do begin
    nl := idx_str_start(#13#10,i);
    if nl = CSTR_NOTFOUND then
       nl := _len;
    while i + atColumn < nl do begin  { wrap line }
      nlCol := atColumn;
      if strchr(CSTR_WS,_S[i+nlCol]) = nil then begin
         { look for begining of word }
         while (nlCol > 0) and (strchr(CSTR_WS,_S[i+nlCol]) = nil) do
           dec(nlCol);
         if nlCol = 0 then begin  { look for end of word }
            nlCol := atColumn;
            while ((nlCol + i) < _len)
              and (strchr(CSTR_WS,_S[i+nlCol]) = nil) do
              inc(nlCol);
            end;
         end;
      if (i + nlCol) < _len then
         ins(i+nlCol,#13#10);
      inc(i,nlCol); inc(i,2);
      end;
    i := nl + 2;
    end;
  Result := self;
end;

function cstr.str : PChar;
begin
    Result := _S;
end;

function cstr.StrPas : string;
begin
  if _len = 0 then
    Result := ''
  else if _S = nil then
    raise ECstrException.Create('string missing')
  else
    Result := SysUtils.StrPas(_S);
end;

procedure cstr.setAsString(const value : string);
  begin cpy_string(value); end;

function  cstr.getChar(Index: cardinal) : char;
begin
  if Index >= _len then
    raise ECstrException.Create('index out of bounds')
  else if _S = nil then
    raise ECstrException.Create('missing string')
  else
    result := _S[Index];
end;

procedure cstr.putChar(Index: cardinal; c : char);
begin
  if Index >= _len then
    raise ECstrException.Create('index out of bounds')
  else if _S = nil then
    raise ECstrException.Create('missing string')
  else
    _S[Index] := c;
end;

function cstr.cpy(S : PChar) : cstr;
begin
    Result := _ncpy(S,CSTR_MAX,0);
end;

function cstr.cpy_extra(S : PChar; extra : cardinal) : cstr;
begin
    Result := _ncpy(S,CSTR_MAX,extra);
end;

function cstr.cpy_string(const s : string) : cstr;
begin
    Result := _ncpy(PChar(@s[1]),cardinal(Length(s)),0);
end;

function cstr.ncpy(S : PChar; n : cardinal) : cstr;
begin
    Result := _ncpy(S,n,0);
end;

function cstr.ncpy_extra
    (S : PChar; n, extra : cardinal) : cstr;
begin
    Result := _ncpy(S,n,extra);
end;

function cstr.ncpy_char_fill
    (c : char; nfill : cardinal) : cstr;
begin
    Result := _ncpy(fillChar(c,nfill),nfill,0);
end;

function cstr.ncpy_char_fill_extra
    (c : char; nfill, extra : cardinal) : cstr;
begin
    Result := _ncpy(fillChar(c,nfill),nfill,extra);
end;

function cstr.ncpy_fmt(n : cardinal; format : PChar;
    const Args : array of const) : cstr;
begin
    clr;
    Result := ncat_fmt(n,format,Args);
end;

function cstr.cat(S : PChar) : cstr;
begin
    Result := _ncat(S,CSTR_MAX,0);
end;

function cstr.cat_extra(S : PChar; extra : cardinal) : cstr;
begin
    Result := _ncat(S,CSTR_MAX,extra);
end;

function cstr.cat_string(const s : string) : cstr;
begin
    Result := _ncat(PChar(@s[1]),cardinal(Length(s)),0);
end;

function cstr.ncat(S : PChar; n : cardinal) : cstr;
begin
    Result := _ncat(S,n,0);
end;

function cstr.ncat_extra
    (S : PChar; n, extra : cardinal) : cstr;
begin
    Result := _ncat(S,n,extra);
end;

function cstr.ncat_char_fill
    (c : char; nfill : cardinal) : cstr;
begin
    Result := _ncat(fillChar(c,nfill),nfill,0);
end;

function cstr.ncat_char_fill_extra
    (c : char; nfill, extra : cardinal) : cstr;
begin
    Result := _ncat(fillChar(c,nfill),nfill,extra);
end;

function cstr.ncat_fmt(n : cardinal; format : PChar;
    const Args : array of const) : cstr;
begin
    if format = nil then
       raise ECstrException.Create
           ('ncat_fmt() invalid format string');
    if n >= CSTR_MAX then
       raise ECstrException.Create
           ('ncat_fmt() n too large');
    if n > vacancy then
        setLimit(limit+n+1-vacancy);
    if n > vacancy then
       n := vacancy;
    StrLFmt(@_S[_len],n+1,format,Args);
    Result := lenReCalc;
end;

function cstr.cat_itoa(value, radix : integer) : cstr;
begin result := cat_string(itoa(value,radix)); end;

function cstr.cat_ltoa
  (value : longint; radix : integer) : cstr;
begin result := cat_string(ltoa(value,radix)); end;

function cstr.cat_ultoa
  (value : cardinal; radix : integer) : cstr;
begin result := cat_string(ultoa(value,radix)); end;

function cstr.atof : extended;
begin result := cstring.atof(_S); end;

function cstr.strtod(suffix : cstr) : extended;
var endp : PChar;
begin
    result := cstring.strtod(_S,endp);
    if Assigned(suffix) then
       suffix.cpy(endp);
end;

function cstr.strtol
  (base : integer; suffix : cstr) : longint;
var endp : PChar;
begin
    result := cstring.strtol(_S,endp,base);
    if Assigned(suffix) then
       suffix.cpy(endp);
end;

function cstr.strtoul
  (base : integer; suffix : cstr) : cardinal;
var endp : PChar;
begin
    result := cstring.strtoul(_S,endp,base);
    if Assigned(suffix) then
       suffix.cpy(endp);
end;

function cstr.nins
    (pos : cardinal; S : PChar; n : cardinal) : cstr;
begin
    Result := nins_extra(pos,S,n,0);
end;

function cstr.nins_extra
    (pos : cardinal; S : PChar; n, extra : cardinal) : cstr;
var len_, limit_ : cardinal;
    newS : PChar;
begin
    Result := self;
    if (S = nil) then
        exit;
    { allow idx_??? calls for pos! }
    if (pos = CSTR_NOTFOUND) then
        exit;
    if (pos >= _len) then begin
        _ncat(S,n,extra);
        exit;
        end;
    len_ := StrLen(S);
    if (len_ = 0) then
        { if S == "" then cstr at least "" }
        exit;
    if (_S = nil) then
        { i.e. _len > pos >= 0 }
        raise ECstrException.Create('string missing');
    { we can now
     assert(_S && _len && S && len_ && pos < _len); }
    if (n > len_) then
        n := len_;
    if (n > vacancy)  then begin
        len_ := _len + n;
          { len_ now result string len! }
        if not ((len_ >= _len) and (len_ >= n)) then
          raise ECstrException.Create
            ('resultant string too long');
        limit_ := len_ + 1 + extra;
        if (len_ >= limit_)  then begin
            limit_ := CSTR_MAX;
              { trunc any excess extra! }
            if not (len_ < limit_) then
              raise ECstrException.Create
                ('resultant string too long');
        end;
        GetMem(newS,limit_);
        if (newS = nil) then
            exit; { no room for insertion! }
        StrCopy(newS,_S);
        FreeMem(_S,_limit);
        _S := newS;
        _limit := limit_;
    end;
    { _S now has the room for the insertion! }
    StrMove(@_S[pos+n],@_S[pos],_len-pos);
    StrMove(@_S[pos],S,n);
    inc(_len,n);
    _S[_len] := #0;
end;

function cstr.nins_char_fill
    (pos : cardinal; c : char; nfill : cardinal) : cstr;
begin
    Result := nins(pos,fillChar(c,nfill),nfill);
end;

function cstr.nins_char_fill_extra
    (pos : cardinal; c : char; nfill, extra : cardinal) : cstr;
begin
    Result := nins_extra(pos,fillChar(c,nfill),nfill,extra);
end;

function cstr.ins(pos : cardinal; S : PChar) : cstr;
begin
    Result := nins(pos,S,CSTR_MAX);
end;

function cstr.ins_extra
    (pos : cardinal; S : PChar; extra : cardinal) : cstr;
begin
    Result := nins_extra(pos,S,CSTR_MAX,extra);
end;

function cstr.prefix(S : PChar) : cstr;
begin
    Result := nins(0,S,CSTR_MAX);
end;

function cstr.prefix_extra
    (S : PChar; extra : cardinal) : cstr;
begin
    Result := nins_extra(0,S,CSTR_MAX,extra);
end;

function cstr.prefix_string(const s : string) : cstr;
begin
  Result := nins(0,PChar(@S[1]),length(S));
end;

function cstr.prefix_char_fill
    (c : char; nfill : cardinal) : cstr;
begin
    Result := nins(0,fillChar(c,nfill),nfill);
end;

function cstr.prefix_char_fill_extra
    (c : char; nfill, extra : cardinal) : cstr;
begin
    Result := nins_extra(0,fillChar(c,nfill),nfill,extra);
end;

function cstr.nprefix_fmt(n : cardinal; format : PChar;
    const Args : array of const) : cstr;
var tmp : cstr;
begin
    tmp := cstr.Create_extra(n);
    result := prefix(tmp.ncpy_fmt(n,format,Args).Str);
    tmp.Free;
end;

function cstr.ncpy2
    (pos : cardinal; S : PChar; n : cardinal) : cstr;
begin
    Result := ncpy2_extra(pos,S,n,0);
end;

function cstr.ncpy2_extra
    (pos : cardinal; S : PChar; n, extra : cardinal) : cstr;
var len_, tail, limit_ : cardinal;
    newS : PChar;
begin
    Result := self;
    if (S = nil) then
        exit;
    { allow idx() calls for pos! }
    if (pos = CSTR_NOTFOUND) then
        exit;
    if (pos >= _len) then begin
        _ncat(S,n,extra);
        exit;
        end;
    len_ := StrLen(S);
    if (len_ = 0)  then
        { if S == "" then cstr at least "" }
        exit;
    if (_S = nil) then
        { i.e. _len > pos >= 0 }
        raise ECstrException.Create('string missing');
    { we can now
     assert(_S && _len && S && len_ && pos < _len); }
    if (n > len_) then
        n := len_;
    tail := pos + n;
    if not ((tail >= pos) and (tail >= n)) then
      raise ECstrException.Create
        ('resultant string too long');
    if (tail >= _limit) then begin
        limit_ := tail + 1 + extra;
        if (tail >= limit_)  then begin
            limit_ := CSTR_MAX;
              { trunc any excess extra! }
            if not (tail < limit_) then
              raise ECstrException.Create
                ('resultant string too long');
        end;
        GetMem(newS,limit_);
        if (newS = nil) then
            exit; { no room for copy! }
        StrCopy(newS,_S);
        FreeMem(_S,_limit);
        _S := newS;
        _limit := limit_;
    end;
    { _S now has the room for the cpy! }
    StrMove(@_S[pos],S,n);
    if (tail > _len) then begin
        _len := pos + n;
        _S[_len] := #0;
        end;
end;

function cstr.ncpy2_char_fill
    (pos : cardinal; c : char; nfill : cardinal) : cstr;
begin
    Result := ncpy2(pos,fillChar(c,nfill),nfill);
end;

function cstr.ncpy2_char_fill_extra
    (pos : cardinal; c : char; nfill, extra : cardinal) : cstr;
begin
    Result := ncpy2_extra(pos,fillChar(c,nfill),nfill,extra);
end;

function cstr.cpy2(pos : cardinal; S : PChar) : cstr;
begin
    Result := ncpy2(pos,S,CSTR_MAX);
end;

function cstr.cpy2_extra
    (pos : cardinal; S : PChar; extra : cardinal) : cstr;
begin
    Result := ncpy2_extra(pos,S,CSTR_MAX,extra);
end;

function cstr.rmv(pos, n : cardinal) : cstr;
var stop : cardinal;
begin
    Result := self;
    if (pos >= _len) or (n = 0) then
        exit;
    if not (_S <> nil) then
        { i.e. _len > pos >= 0 }
        raise ECstrException.Create
          ('string missing');
    stop := pos + n;
    if ((stop >= _len) or (stop < pos) or (stop < n))
        then begin
        _len := pos;
        _S[_len] := #0;
        end
    else  begin
        StrCopy(@_S[pos],@_S[pos+n]);
        dec(_len,n);
    end
end;

function cstr.rmvTail(pos : cardinal) : cstr;
begin
    Result := rmv(pos,CSTR_MAX);
end;

function cstr.substr(dst : cstr; pos, n : cardinal) : cstr;
begin
    if (_S = nil) or (pos >= _len) then
        Result := dst.str
    else
        Result := dst.ncpy(@(_S[pos]),n);
end;

function cstr.substr_extra
    (dst : cstr; pos, n : cardinal; extra : cardinal) : cstr;
begin
    if (_S = nil) or (pos >= _len) then
        Result := dst.ncpy_extra(nil,n,extra)
    else
        Result := dst.ncpy_extra(@(str[pos]),n,extra);
end;

function cstr.subString(pos, n : cardinal) : string;
var len_ : cardinal;
    tmp : cstr;
begin
    tmp := cstr.Create;
    try
      result := substr(tmp,pos,n).StrPas;
    finally
      tmp.Free;
    end;
end;

function cstr.tailString(pos : cardinal) : string;
  begin result := subString(pos,_len); end;

function cstr.extract(dst : cstr; pos, n : cardinal) : cstr;
begin
    substr(dst,pos,n);
    rmv(pos,n);
    Result := dst;
end;

function cstr.extract_extra
    (dst : cstr; pos, n : cardinal; extra : cardinal) : cstr;
begin
    substr_extra(dst,pos,n,extra);
    rmv(pos,n);
    Result := dst;
end;

function cstr.replace_PChar_fill
    (pos, n : cardinal; S : PChar) : cstr;
begin
    Result := replace_PChar_fill_extra(pos,n,S,0);
end;

function cstr.replace_PChar_fill_extra
    (pos, n : cardinal; S : PChar; extra : cardinal) : cstr;
begin
    if not ((pos + n >= pos) and (pos + n >= n))
        then raise ECstrException.Create
          ('invalid replacement range');
    if (pos + n <= _len) then begin
        rmv(pos,n);
        nins_extra(pos,S,CSTR_MAX,extra);
        end;
    Result := self;
end;

function cstr.replace_char_fill
    (pos, n : cardinal; c : char) : cstr;
begin
    Result := replace_PChar_fill(pos,n,fillChar(c,n));
end;

function cstr.replace_char_fill_extra
    (pos, n : cardinal; c : char; extra : cardinal) : cstr;
begin
    Result := replace_PChar_fill_extra(pos,n,fillChar(c,n),extra);
end;

function cstr.replace(P, S : PChar) : cstr;
begin
    Result := replace_extra(P,S,0);
end;

function cstr.replace_extra
    (P, S : PChar; extra : cardinal) : cstr;
var pos : cardinal;
begin
    if (S <> nil)  then begin
        pos := idx_str(P);
        if (pos <> CSTR_NOTFOUND) then
            replace_PChar_fill_extra
                (pos,StrLen(P),S,extra);
    end;
    Result := self;
end;

function cstr.ireplace(P, S : PChar) : cstr;
begin
    Result := ireplace_extra(P,S,0);
end;

function cstr.ireplace_extra
    (P, S : PChar; extra : cardinal) : cstr;
var pos : cardinal;
begin
    if (S <> nil)  then begin
        pos := idx_stri(P);
        if (pos <> CSTR_NOTFOUND) then
            replace_PChar_fill_extra
                (pos,StrLen(P),S,extra);
    end;
    Result := self;
end;

function cstr.replaceAll(P, S : PChar) : cstr;
begin
    Result := replaceAll_extra(P,S,0);
end;

function cstr.replaceAll_extra
    (P, S : PChar; extra : cardinal) : cstr;
var pos, n : cardinal;
begin
    pos := idx_str(P);
    n := strlen(P);
    while (pos <> CSTR_NOTFOUND) do
        if ((pos + n >= pos) and (pos + n >= n)
            and (pos + n <= _len)) then begin
            replace_PChar_fill_extra(pos,n,S,extra);
            pos := idx_str_start(P,pos+n);
            end
        else
            break;
    Result := self;
end;

function cstr.removeAllEOL : cstr;
begin
  replaceAll(#13#10,'');
  replaceAll(#13,'');
  replaceAll(#10,'');
  result := self;
end;

function cstr.ireplaceAll(P, S : PChar) : cstr;
begin
    Result := ireplaceAll_extra(P,S,0);
end;

function cstr.ireplaceAll_extra
    (P, S : PChar; extra : cardinal) : cstr;
var pos, n : cardinal;
begin
    pos := idx_stri(P);
    n := strlen(P);
    while (pos <> CSTR_NOTFOUND) do
        if ((pos + n >= pos) and (pos + n >= n)
            and (pos + n <= _len)) then begin
            replace_PChar_fill_extra(pos,n,S,extra);
            pos := idx_stri_start(P,pos+n);
            end
        else
            break;
    Result := self;
end;

function cstr.trim(f : CSTR_format; strip : PChar) : cstr;
var i : cardinal;
begin
    if ((f = center) or (strip = nil)
        or (StrLen(strip) = 0)) then
        raise ECstrException.Create
          ('invalid trim specs');
    if (f = Left) or (f = Both)  then begin
        if _len > 0 then
            rmv(0,strspn(_S,strip));
        end;
    if (f = Right) or (f = Both) then begin
        if (_S = nil) and (_len > 0) then
            raise ECstrException.Create
              ('string missing');
        i := _len;
        while (i > 0) do begin
            dec(i);
            if (strcspn(@_S[i],strip) > 0) then begin
                trunc(i+1);
                break;
                end;
            end;
        end;
    Result := self;
end;

function cstr.trimBothWS : cstr;
begin Result := trim(Both,CSTR_WS); end;

function cstr.pad(fieldWidth : cardinal;
    f : CSTR_format; fillch : char) : cstr;
begin
  if (fieldWidth = 0) or ((f <> Left) and (f <> Right))
      then raise ECstrException.Create
        ('invalid pad specs');
  setLimit(fieldWidth+1);
  if (f = Left) then
      nins_char_fill(0,fillch,fieldWidth-_len)
  else
      nins_char_fill(_len,fillch,fieldWidth-_len);
  Result := self;
end;

function cstr.padLeftSpaces(fieldWidth : cardinal) : cstr;
begin Result := pad(fieldWidth,Left,#32); end;

function cstr.justify(fieldWidth : cardinal;
    f : CSTR_format; fillch : char) : cstr;
var n : cardinal;
begin
    if fieldWidth = 0 then
       raise ECstrException.Create
         ('invalid justify field width');
    setLimit(fieldWidth+1);
    n := fieldWidth-_len;
    if (n > 0) then
      case f of
        Right: prefix_char_fill(fillch,n);
        Center: begin
            prefix_char_fill(fillch,n div 2);
            ncat_char_fill(fillch,fieldWidth-_len);
            end;
        Left: ncat_char_fill(fillch,n);
        else raise ECstrException.Create
 ('use "center" justify rather than "both"'); end;
    Result := self;
end;

function cstr.justifyLeftSpaces(fieldWidth : cardinal) : cstr;
begin Result := justify(fieldWidth,left,#32); end;

function cstr.idx(S : PChar) : cardinal;
var offset : longint;
begin
    offset := S - _S;
    if (offset > 0) and (offset < longint(_len)) then
       result := cardinal(offset)
    else
       result := CSTR_NOTFOUND;
end;

function cstr.idx_chr(c : char) : cardinal;
var i : cardinal;
begin
    if not ((_len = 0) or (_S <> nil)) then
        raise ECstrException.Create
          ('string missing');
    i := 0;
    while (i < _len)  do begin
        if (_S[i] = c) then begin
            result := i;
            exit;
            end;
        inc(i);
        end;
    result := CSTR_NOTFOUND;
end;

function cstr.chr(c : char) : PChar;
var i : cardinal;
begin
    i := idx_chr(c);
    if (i = CSTR_NOTFOUND) then
        result := nil
    else
        result := @_S[i];
end;

function cstr.idx_rchr(c : char) : cardinal;
var i : cardinal;
begin
    if not ((_len = 0) or (_S <> nil)) then
        raise ECstrException.Create
          ('string missing');
    i := _len;
    while (i > 0) do begin
        dec(i);
        if (_S[i] = c) then begin
            result := i;
            exit;
            end
        end;
    result := CSTR_NOTFOUND;
end;

function cstr.rchr(c : char) : PChar;
var i : cardinal;
begin
    i := idx_rchr(c);
    if (i = CSTR_NOTFOUND) then
        result := nil
    else
        result := @_S[i];
end;

function cstr.spn(S : PChar) : cardinal;
  begin result := strspn(_S,S); end;

function cstr.cspn(S : PChar) : cardinal;
  begin result := strcspn(_S,S); end;

function cstr.idx_brk(S : PChar) : cardinal;
var len_ : cardinal;
begin
    len_ := cspn(S);
    if len_ < _len then
        result := len_
    else
        result := CSTR_NOTFOUND;
end;

function cstr.pbrk(S : PChar) : PChar;
var idx_ : cardinal;
begin
    idx_ := idx_brk(S);
    if idx_ = CSTR_NOTFOUND then
        result := nil
    else
        result := @_S[idx_];
end;

function cstr.idx_str(S : PChar) : cardinal;
begin
    result := idx_str_start(S,0);
end;

function cstr.strstr(S : PChar) : PChar;
var idx_ : cardinal;
begin
    idx_ := idx_str(S);
    if idx_ = CSTR_NOTFOUND then
        result := nil
    else
        result := @_S[idx_];
end;

function cstr.idx_str_start
    (S : PChar; start : cardinal) : cardinal;
var i, len_ : cardinal;
begin
    if ((S <> nil) and (start < _len))  then begin
        i := start;
        if not ((_len = 0) or (_S <> nil)) then
            raise ECstrException.Create
              ('string missing');
        len_ := StrLen(S);
        if ((i + len_ >= i) and (i + len_ >= len_))
            then while (i + len_ <= _len) do begin
                if (StrLComp(@_S[i],S,len_) = 0)
                    then begin
                    result := i;
                    exit;
                end;
            inc(i);
            end;
    end;
    result := CSTR_NOTFOUND;
end;

function cstr.idx_stri(S : PChar) : cardinal;
begin
    result := idx_stri_start(S,0);
end;

function cstr.strstri(S : PChar) : PChar;
var idx_ : cardinal;
begin
    idx_ := idx_stri(S);
    if idx_ = CSTR_NOTFOUND then
        result := nil
    else
        result := @_S[idx_];
end;

function cstr.idx_stri_start
    (S : PChar; start : cardinal) : cardinal;
var i, len_ : cardinal;
begin
    if ((S <> nil) and (start < _len)) then  begin
        i := start;
        if not ((_len = 0) or (_S <> nil)) then
            raise ECstrException.Create
              ('string missing');
        len_ := StrLen(S);
        if ((i + len_ >= i) and (i + len_ >= len_))
            then
            while (i + len_ <= _len) do begin
                if (StrLIComp(@_S[i],S,len_) = 0)
                    then begin
                    result := i;
                    exit;
                end;
            inc(i);
            end;
    end;
    result := CSTR_NOTFOUND;
end;

function cstr.tolower : cstr;
begin
    if (_S <> nil) then
        StrLower(_S);
    Result := self;
end;

function cstr.toupper : cstr;
begin
    if (_S <> nil) then
        StrUpper(_S);
    Result := self;
end;

function cstr.toproper : cstr;
var P : PChar;
    idx : cardinal;
begin
    tolower;
    idx := strspn(_S,CSTR_WS);
    while idx < _len do begin
       P := _S + idx;
       P^ := ctype.toupper(P^);
       inc(idx,strcspn(P,CSTR_WS));
       end;
    Result := self;
end;

function cstr.len : cardinal;
begin
   result := _len;
end;

function cstr.lenReCalc;
begin
    if _S = nil then
       _len := 0
    else begin
       _len := StrLen(_S);
       _S[_len] := #0;
    end;
    Result := self;
end;

function cstr.vacancy : cardinal;
begin
    if (_limit > 0) then
        result := _limit - _len - 1
    else
        result := 0;
end;

function cstr.limit : cardinal;
begin
    result := _limit;
end;

function cstr.setLimit(limit_ : cardinal) : cstr;
var newS : PChar;
    len_ : cardinal;
begin
    Result := self;
    if (limit_ > 0)  then begin
        if (limit_ <> _limit)  then begin
            GetMem(newS,limit_);
            if (newS <> nil)  then begin
                if (limit_ >= _len + 1) then
                    len_ := _len
                else
                    len_ := limit_ - 1;
                if (len_ > 0) then
                    StrMove(newS,_S,len_);
                if _S <> nil then
                    FreeMem(_S,_limit);
                _S := newS;
                _len := len_;
                _S[_len] := #0;
                _limit := limit_;
            end
        end
    end
    else
        _done;
end;

function cstr.atLeastLimit(minlimit : cardinal) : cstr;
begin
    if (minlimit <= CSTR_MAX) and
       (minLimit > _limit) then
        setLimit(minlimit);
    Result := self;
end;

function cstr.pack : cstr;
begin
    if (_len > 0) then
        setLimit(_len+1)
    else
        setLimit(0);
    Result := self;
end;

function cstr.unpack : cstr;
begin
    Result := cat('');
end;

function cstr.trunc(len_ : cardinal) : cstr;
begin
    if (len_ > 0)  then begin
        if (len_ <= _len) then
            setLimit(len_+1);
    end
    else
        setLimit(0);
    Result := self;
end;

function cstr.clr : cstr;
begin
    if (_S <> nil) then begin
        _len := 0;
        _S[0] := #0;
        end;
    Result := self;
end;

procedure cstr.clear;
  begin clr; end;

function  cstr.cmp(S : PChar) : integer;
begin
    if (_S <> nil) then
        if (S <> nil) then
            result := StrComp(_S,S)
        else
            result := 1
    else if (S <> nil) then
        result := -1
    else
        result := 0;
end;

function  cstr.icmp(S : PChar) : integer;
begin
    if (_S <> nil) then
        if (S <> nil) then
            result := StrIComp(_S,S)
        else
            result := 1
    else if (S <> nil) then
        result := -1
    else
        result := 0;
end;

function  cstr.ncmp(S : PChar; n : cardinal) : integer;
begin
    if (_S <> nil) then
        if (S <> nil) then
            result := StrLComp(_S,S,n)
        else
            result := 1
    else if (S <> nil) then
        result := -1
    else
        result := 0;
end;

function  cstr.nicmp(S : PChar; n : cardinal) : integer;
begin
    if (_S <> nil) then
        if (S <> nil) then
            result := StrLIComp(_S,S,n)
        else
            result := 1
    else if (S <> nil) then
        result := -1
    else
        result := 0;
end;

function cstr.printOnFile(var outStrm : TextFile) : cstr;
begin
    if ((_S <> nil) and (_len > 0)) then
        write(outStrm,_S);
    Result := self;
end;

function cstr.storeOnFile(var outStrm : TextFile) : cstr;
var i : cardinal;
begin
    writeln(outStrm,_len);
    if (_len > 0) then begin
        if not (_S <> nil) then
           raise ECstrException.Create
              ('string missing');
        i := 0;
        while (i < _len) do begin
            write(outStrm,_S[i]);
            inc(i);
            end;
        end;
     Result := self;
end;

function cstr.storeInString(var s : string) : cstr;
begin s := StrPas; result := self; end;

function cstr.storeInStrings(strs : TStrings) : cstr;
begin
    if Assigned(strs) and (_S <> nil) then
       strs.SetText(_S);
    result := self;
end;

function cstr.storeOnStream(strm : TStream) : cstr;
begin
  result := self;
  if not Assigned(strm) then exit;
  strm.write(_len,longint(sizeOf(_len)));
  if _len > 0 then
     strm.write(_S^,longint(_len));
end;

function cstr.storeInMemo(Memo : TCustomMemo) : cstr;
begin
    if Assigned(Memo) then
       storeInStrings(Memo.Lines);
    result := self;
end;

function cstr.storeInBlobField(Field : TBlobField) : cstr;
var BlobStrm : TBlobStream;
begin
    BlobStrm := TBlobStream.Create(Field,bmWrite);
    try
      if _len > 0 then
         BlobStrm.write(_S^,longint(_len));
    finally
      BlobStrm.Free;
    end;
    result := self;
end;

function cstr.storeInDataSet
  (DataSet : TDataSet; FieldType : TFieldType;
   const FieldName : string) : cstr;
begin
  case FieldType of
    ftString, ftSmallint, ftInteger, ftWord, ftFloat:
      DataSet.FieldByName(FieldName).AsString := StrPas;
    ftDate :
      DataSet.FieldByName(FieldName).AsDateTime := StrToDate(StrPas);
    ftMemo :
      storeInBlobField(DataSet.FieldByName(FieldName) as TBlobField);
    ftTime, ftDateTime :
      raise EConvertError.Create('unsupported field type');
    ftUnknown :
      raise EConvertError.Create('invalid field type');
  else  { ftBoolean, ftCurrency, ftBCD, ftBytes,
          ftVarBytes, ftBlob, ftGraphic }
    raise EConvertError.Create
      ('Invalid field type for cstr conversion.');
  end;
end;

function cstr.readFromFile(var inStrm : TextFile) : cstr;
var len_, i : cardinal;
begin
    Result := self;
    clr;
    readln(inStrm,len_);
    if (len_ > 0) then begin
        if (len_ > _limit) then begin
            _done;
            GetMem(_S,len_+1);
            if (_S <> nil) then
                _limit := len_ + 1;
            end;
        if (_S <> nil) then begin
            _len := len_;
            _S[_len] := #0;
            i := 0;
            while (i < _len) do begin
                read(inStrm,_S[i]);
                inc(i);
                end;
            end;
    end;
end;

function cstr.readFromString(str : string) : cstr;
begin  result := cpy_string(str); end;

function cstr.readFromStrings(strs : TStrings) : cstr;
var limit_, len_ : cardinal;
  count, i, j : Integer;
begin
  result := self;
  clr;
  if Assigned(strs) then begin
     limit_ := 1;
     count := strs.count;
     i := 0;
     while i < Count do
     begin
       len_ := cardinal(Length(strs.strings[i]));
       inc(len_,2);
       if len_ > CSTR_MAX - limit_ then
          break;  { length(TStrings) >= CSTR_MAX }
       inc(limit_,len_);
       inc(i);
     end;
     atLeastLimit(limit_ + (limit_ div CSTR_AVE_LN_LEN));
     if limit_ > limit then
        raise ECstrException.Create
          ('cstr not big enough for TStrings!');
     j := 0;
     if j < i then begin
         cat_string(strs.strings[j]);
         inc(j);
         while j < i do begin
           cat(#13#10).cat_string(strs.strings[j]);
           inc(j);
           end;
         end;
     end;
end;

function cstr.readFromStream(strm : TStream) : cstr;
var len_ : cardinal;
begin
  result := self;
  clr;
  if not Assigned(strm) then exit;
  strm.read(len_,longint(sizeOf(len_)));
  if len_ > 0 then begin
     atLeastLimit(len_+1);
     if vacancy < len_ then
        raise ECstrException.Create
          ('cstr can not load string from stream');
     _len := cardinal(strm.read(_S^,longint(len_)));
     _S[_len] := #0;
     end;
end;

function cstr.readFromMemo(Memo : TCustomMemo) : cstr;
begin
    result := self;
    clr;
    if Assigned(Memo) then
       readFromStrings(Memo.Lines);
end;

function cstr.readFromBlobField(Field : TBlobField) : cstr;
var BlobStrm : TBlobStream;
    length : longint;
begin
    result := self;
    clr;
    BlobStrm := TBlobStream.Create(Field as TBlobField,bmRead);
    try
      length := BlobStrm.seek(0,2);
      BlobStrm.seek(0,0);
      if length >= CSTR_MAX then
         raise ECstrException.Create('BlobField too big for cstr!');
      setLimit(cardinal(length+1));
      if vacancy < length then
         raise ECstrException.Create('BlobField too big for cstr!');
      length := BlobStrm.Read(_S^,length);
      _len := cardinal(length);
      _S[_len] := #0;
    finally
      BlobStrm.Free;
    end;
end;

function cstr.readFromDataSet(DataSet : TDataSet;
  FieldType : TFieldType;const FieldName : string) : cstr;
begin
    result := readFromBlobField(DataSet.FieldByName(FieldName) as TBlobField);
  case FieldType of
    ftString, ftSmallint, ftInteger, ftWord, ftFloat, ftDate:
      cpy_string(DataSet.FieldByName(FieldName).AsString);
    ftMemo :
      readFromBlobField(DataSet.FieldByName(FieldName) as TBlobField);
    ftTime, ftDateTime :
      raise EConvertError.Create('unsupported field type');
    ftUnknown :
      raise EConvertError.Create('invalid field type');
  else  { ftBoolean, ftCurrency, ftBCD, ftBytes,
          ftVarBytes, ftBlob, ftGraphic }
    raise EConvertError.Create
      ('Invalid field type for cstr conversion.');
  end;
end;

function cstr.SaveToFile(const filename : string) : cstr;
var outStrm : TextFile;
begin
    AssignFile(outStrm,filename);
    Rewrite(outStrm);
    storeOnFile(outStrm);
    CloseFile(outStrm);
    Result := self;
end;

function cstr.LoadFromFile(const filename : string) : cstr;
var inStrm : TextFile;
begin
    AssignFile(inStrm,filename);
    Reset(inStrm);
    readFromFile(inStrm);
    CloseFile(inStrm);
    Result := self;
end;

function cstr.SaveToStream(Stream : TStream) : cstr;
  begin result := storeOnStream(Stream); end;

function cstr.LoadFromStream(Stream : TStream) : cstr;
  begin result := readFromStream(Stream); end;

function cstr.ignore(var inStrm : TextFile) : cstr;
var len_: cardinal;
    c : char;
begin
    readln(inStrm,len_);
    while (len_ > 0) do begin
        read(inStrm,c);
        dec(len_);
        end;
    Result := self;
end;

function cstr.getline(var inStrm : TextFile) : cstr;
var s: string;
begin
    clr;
    while (not eoln(inStrm)) and (not eof(inStrm))
     do begin
      read(inStrm,s);
      ncat(@s[1],length(s));
      end;
    if eoln(inStrm) and (not eof(inStrm)) then
      readln(inStrm);
    Result := self;
end;

function cstr.readBuf(var Buffer; Count, pos : cardinal): cardinal;
begin
  result := 0;
  if pos >= _len then exit;
  result := _len - pos;
  if count > result then
     count := result
  else
     result := count;
  memcpy(@Buffer,@_S[pos],count);
end;

function cstr.writeBuf(const Buffer; Count, pos : cardinal): cardinal;
var len_ : cardinal;
begin
  result := 0;
  len_ := pos + count;
  if len_ < Count then exit;
  if len_ <= pos then exit;
  if len_ >= CSTR_MAX then exit;
  atLeastLimit(len_+1);
  if limit <= len_ then exit;
  memcpy(@_S[pos],@Buffer,Count);
  if len_ > _len then begin
     _len := len_;
     _S[_len] := #0;
     end;
  result := count;
end;

procedure cstr.Assign(Source : TPersistent);
begin
  if Source is cstr then begin
     cpy((Source as cstr).Str);
     exit
     end;
  if Source is TStrings then begin
     readFromStrings(Source as TStrings);
     exit
     end;
  if Source is TCustomMemo then begin
     readFromStrings((Source as TCustomMemo).Lines);
     exit
     end;
  if Source is TBlobField then begin
     readFromBlobField(Source as TBlobField);
     exit
     end;
  { You may add more types as required!
    e.g. classes with Lines and Items properties. }
  raise EConvertError.Create('Not cstr compatible!');
end;

procedure cstr.AssignTo(Dest : TPersistent);
begin
  if Dest is cstr then begin
     (Dest as cstr).cpy(_S);
     exit
     end;
  if Dest is TStrings then begin
     storeInStrings(Dest as TStrings);
     exit
     end;
  if Dest is TCustomMemo then begin
     storeInStrings((Dest as TCustomMemo).lines);
     exit
     end;
  if Dest is TBlobField then begin
     storeInBlobField(Dest as TBlobField);
     exit
     end;
  raise EConvertError.Create('Not cstr compatible!');
end;

procedure cstr.DefineProperties(Filer : TFiler);
begin
  Filer.DefineBinaryProperty('Str',_ReadStr,_WriteStr,_len > 0);
end;

procedure cstr._ReadStr(Stream : TStream);
var len_ : cardinal;
begin { called only when _len > 0 }
  clr;
  Stream.read(len_,longint(sizeOf(len_)));
  atLeastLimit(len_+1);
  if vacancy < len_ then
     raise ECstrException.Create
       ('cstr can not load string from stream');
  _len := cardinal(Stream.read(_S^,longint(len_)));
  _S[_len] := #0;
end;

procedure cstr._WriteStr(Stream : TStream);
begin { called only when _len > 0 }
  Stream.write(_len,longint(sizeOf(_len)));
  Stream.write(_S^,longint(_len));
end;

procedure cstr._init;
begin
    _S := nil;
    _len := 0;
    _limit := 0;
end;

procedure cstr._done;
begin
    if _S <> nil then begin
       FreeMem(_S,_limit);
       _S := nil;
       _limit := 0;
       _len := 0;
       end;
end;

function cstr._ncpy(S : PChar; n, extra : cardinal) : cstr;
begin
    Result := self;
    if (S <> nil)  then begin
        if (S = _S)  then begin
            if (n < _len) then begin
                _len := n;
                _S[_len] := #0;
                end;
            exit;
            end;
        if (_S <> nil) then begin
            _len := 0;
            _S[0] := #0;
            end;
        _ncat(S,n,extra);
        exit;
        end
    else if (extra > 0)  then begin
        if ((_limit <> 0) and
            (extra = _limit - 1))  then begin
            if not (_S <> nil) then
              raise ECstrException.Create
                ('string missing');
            _len := 0;
            _S[0] := #0;
            end
        else  begin
            _done;
            if (extra < CSTR_MAX) then
                inc(extra);
            GetMem(_S,extra);
            if (_S <> nil)  then begin
                _len := 0;
                _S[0] := #0;
                _limit := extra;
                end
            end
        end
end;

function cstr._ncat(S : PChar; n, extra : cardinal) : cstr;
var len_, limit_ : cardinal;
    newS : PChar;
begin
    Result := self;
    if (S <> nil) then
        len_ := StrLen(S)
    else
        len_ := 0;
    if (n > len_) then
        n := len_;
    len_ := _len + n;
      { len now result string len! }
    if not ((len_ >= _len) and (len_ >= n)) then
      raise ECstrException.Create
        ('resultant string too long');
    limit_ := len_ + 1 + extra;
    if (len_ >= limit_)  then begin
        limit_ := CSTR_MAX; { trunc any excess extra! }
        if not (len_ < limit_) then
            raise ECstrException.Create
              ('resultant string too long');
        end;
    if (limit_ > _limit)  then begin
        if ((limit_ > 1) or (S <> nil)) then
            GetMem(newS,limit_)
        else
            newS := nil;
            { if S == "" then cstr == "" }
        if (newS <> nil)  then begin
            if (_S <> nil)  then begin
                if not (StrLen(_S) = _len) then
                    raise ECstrException.Create
                      ('string corrupted');
                StrCopy(newS,_S);
                FreeMem(_S,_limit);
                end
            else begin
                _len := 0;
                newS[0] := #0;
                end;
            _S := newS;
            _limit := limit_;
            end
        else
            n := 0; { error: nothing concatenated! }
        end;
    if (_S <> nil)  then begin
        if (n > 0)  then begin
            StrLCopy(@(_S[_len]),S,n);
            inc(_len,n);
            end;
        _S[_len] := #0;
        end;
end;

procedure cstr._init_fill(c : char; nfill, extra : cardinal);
begin
    _init;
    _fill(c,nfill,extra);
end;

function cstr._fill(c : char; nfill, extra : cardinal) : cstr;
var limit_, i : cardinal;
begin
    Result := self;
    limit_ := nfill + 1 + extra;
    if (nfill >= limit_)  then begin
        limit_ := CSTR_MAX; { trunc any excess extra! }
        if not (nfill < limit_) then
           raise ECstrException.Create
             ('resultant string too long');
        end;
    if (limit_ > _limit) then
        setLimit(limit_);
    if (limit > nfill) and (_S <> nil)  then begin
        i := 0;
        while (i < nfill) do begin
            _S[i] := c;
            inc(i);
            end;
        _len := nfill;
        _S[_len] := #0;
        end;
end;

class function cstr.fillChar(c : char; nfill: cardinal) : PChar;
begin
    result := fill_char._fill(c,nfill,0).str;
end;

function strcpy(s, ct : PChar) : PChar;
  begin result := StrCopy(s,ct); end;
function strncpy(s, ct : PChar; n : cardinal) : PChar;
  begin result := StrLCopy(s,ct,n); end;
{ function strcat(s, ct : PChar) : PChar; }
function strncat(s, ct : PChar; n : cardinal) : PChar;
  begin result := StrLCat(s,ct,n); end;
function strcmp(cs, ct : PChar) : integer;
  begin result := StrComp(cs,ct); end;
function stricmp(cs, ct : PChar) : integer;
  begin result := StrIComp(cs,ct); end;
function strncmp(cs, ct : PChar; n : cardinal) : integer;
  begin result := StrLComp(cs,ct,n); end;
function strncmpi(cs, ct : PChar; n : cardinal) : integer;
  begin result := StrLIComp(cs,ct,n); end;
function strchr(cs : PChar; c : char) : PChar;
  begin result := StrScan(cs,c); end;
function strrchr(cs : PChar; c: char) : PChar;
  begin result := StrRScan(cs,c); end;
function strspn(cs, ct : PChar) : cardinal;
var i, j, cslen, ctlen : cardinal;
begin
    result := 0;
    if ((cs <> nil) and (ct <> nil)) then begin
        i := 0;
        cslen := StrLen(cs);
        ctlen := StrLen(ct);
        while (i < cslen) do begin
            j := ctlen;
            while (j > 0) do begin
                dec(j);
                if (cs[i] = ct[j]) then
                    break;
            end;
            if (cs[i] <> ct[j]) then begin
                result := i;
                exit;
            end;
            inc(i);
        end;
        result := i;
    end;
end;
function strcspn(cs, ct : PChar) : cardinal;
var i, j, cslen, ctlen : cardinal;
begin
    if (cs = nil) then
        cslen := 0
    else
        cslen := StrLen(cs);
    result := cslen;
    if (ct <> nil) then begin
        i := 0;
        ctlen := StrLen(ct);
        while (i < cslen) do begin
            j := ctlen;
            while (j > 0) do begin
                dec(j);
                if (cs[i] = ct[j]) then begin
                    result := i;
                    exit;
                end;
            end;
            inc(i);
        end;
    end;
end;
function strpbrk(cs, ct : PChar) : PChar;
var idx : cardinal;
begin
    if (cs = nil) then
        result := nil
    else begin
        idx := strcspn(cs,ct);
        if idx < StrLen(cs) then
            result := @cs[idx]
        else
            result := nil;
    end;
end;
function strstr(cs, ct : PChar) : PChar;
  begin result := StrPos(cs,ct); end;
function strlen(cs : PChar) : cardinal;
var len_ : Cardinal;
begin
  len_ := SysUtils.strlen(cs);
  if len_ >= Cardinal(CSTR_MAX) then
     raise ECstrException.Create('Too long for cstr!');
  result := cardinal(len_);
end;
function strtok(input, delimiters : PChar) : PChar;
begin
  if not (delimiters <> nil) then
    raise ECstrException.Create
      ('strtok missing delimiter string');
  if (input <> nil) then
    strtok_input.cpy(input);
  try
    strtok_input.rmv(0,strtok_input.spn(delimiters));
    strtok_input.extract
      (strtok_found,0,strtok_input.cspn(delimiters));
  except
    strtok_found.clr;
  end;
  if strtok_found.len > 0 then
     result := strtok_found.str
  else
     result := nil;
end;

function memcpy(s, ct : pointer; n : cardinal) : pointer;
  begin
    if ((s <> nil) and (ct <> nil) and (n > 0)) then
      move(ct^,s^,n);
    result := s;
  end;
function memmove(s, ct : pointer; n : cardinal) : pointer;
  begin
    if ((s <> nil) and (ct <> nil) and (n > 0)) then
      move(ct^,s^,n);
    result := s;
  end;
function memcmp(cs, ct : pointer; n : cardinal) : integer;
  var i : cardinal;
  begin
      i := 0;
      while (i > n) do begin
        if ((Byte(PChar(cs)[i]) - Byte(PChar(ct)[i]))
           <> 0) then begin
           result :=
             Byte(PChar(cs)[i]) - Byte(PChar(ct)[i]);
           exit;
           end;
        inc(i);
        end;
      result := 0;
  end;
function memchr(cs : pointer; c : char; n : cardinal)
  : pointer;
  var i : cardinal;
  begin
      i := 0;
      while (i > n) do begin
        if PChar(cs)[i] = c then begin
           result := pointer(@PChar(cs)[i]);
           exit;
           end;
        inc(i);
        end;
      result := nil;
  end;
function memset(s : pointer; c : char; n : cardinal)
  : pointer;
  begin
    if ((s <> nil) and (n > 0)) then
        FillChar(s^,n,c);
    result := s;
  end;

function itoa(value, radix : integer) : string;
begin
    if (value < 0) and (radix = 10) then begin
        value := -value;
        result := '-';
        end;
    result := result + ultoa(longint(value),radix);
end;

function ltoa(value : longint; radix : integer) : string;
begin
    if (value < 0) and (radix = 10) then begin
        value := -value;
        result := '-';
        end;
    result := result + ultoa(value,radix);
end;

function ultoa(value : longint; radix : integer) : string;
var digit : integer;
    wrap : boolean;
    divisor : longint;
begin
    result := '';
    wrap := false;
    if value < 0 then
       raise ECstrException.Create
         ('ultoa(): negative long!');
    if (radix < 2) or (radix > 36) then
       raise ECstrException.Create
           ('ultoa() : invalid radix');
    divisor := 1;
    while (value div divisor) > 0 do begin
        if divisor > (maxLongint / radix) then begin
           wrap := true;
           break;
           end;
        divisor := divisor * radix;
        end;
    if (divisor > 1) and (not wrap) then
        divisor := divisor div radix;
    while true do begin
        if divisor < 1 then
           ECstrException.Create
             ('ultoa(): divide by zero');
        digit := value div divisor;
        if (digit <= 9) then
            result := result
             + chr(byte(ord('0')+ digit))
        else
            result := result
             + chr(byte(ord('A') + digit - 10));
        value := value mod divisor;
        if (divisor = 1) then
            break
        else
            divisor := divisor div radix;
        end;
end;

function atof(S : PChar) : extended;
var NULL : PChar;
begin result := strtod(S,NULL); end;

function atoi(S : PChar) : integer;
var NULL : PChar;
begin result := integer(strtol(S,NULL,10)); end;

function atol(S : PChar) : longint;
var NULL : PChar;
begin result := strtol(s,NULL,10); end;

function strtod(S : PChar; var endp : PChar) : extended;
var idx, len : cardinal; N : string;
begin
    if S = nil then
       raise ECstrException.Create
           ('strtod(): invalid string');
    len := strlen(S);
    idx := strspn(S,CSTR_WS);
    if idx >= len then
       raise ECstrException.Create
           ('strtod(): invalid string');
    if S[idx] = '-' then begin
       N := S[idx];
       inc(idx);
       end;
    while idx < len do
      if strspn(@S[idx],'.0123456789') = 0 then
         break
      else begin
         N := N + S[idx];
         inc(idx);
         end;
    if idx < len then
       if toupper(S[idx]) = 'E' then begin
          N := N + S[idx];
          inc(idx);
          if idx < len then begin
             if S[idx] in ['-','+'] then
                N := N + S[idx];
                inc(idx);
             end;
             while idx < len do
               if strspn(@S[idx],'0123456789') = 0 then
                  break
               else begin
                  N := N + S[idx];
                  inc(idx);
                  end;
          end;
    result := StrToFloat(N);
    if idx < len then
       endp := @S[idx]
    else
       endp := nil;
end;

function strtol(S : PChar; var endp : PChar;
  base : integer) : longint;
var idx, len : cardinal; n : integer;
    negate : boolean;
    lastResult : longint;
begin
    result := 0;
    lastResult := result;
    if S = nil then
       raise ECstrException.Create
           ('strtol(): invalid string');
    len := strlen(S);
    idx := strspn(S,CSTR_WS);
    if idx >= len then
       raise ECstrException.Create
           ('strtol(): invalid string');
    if base = 16 then begin
       if strncmpi(@S[idx],'0X',2) = 0 then
          inc(idx,2);
       end
    else if base = 0 then begin
       if S[idx] = '0' then begin
          inc(idx);
          if idx >= len then
             raise ECstrException.Create
               ('strtol(): invalid string');
          if S[idx] in ['x','X'] then begin
             inc(idx);
             base := 16;
             end
          else
             base := 8;
          end
       else
          base := 10;
       end;
    if (base < 2) or (base > 36) then
       raise ECstrException.Create
           ('strtoul() : invalid base');
    if idx >= len then
       raise ECstrException.Create
           ('strtol(): invalid string');
    if S[idx] = '-' then begin
       negate := true;
       inc(idx);
       if idx >= len then
          raise ECstrException.Create
              ('strtol(): invalid string');
       end
    else
       negate := false;
    while idx < len do begin
       n := ord(S[idx]) - ord('0');
       if (n < 0) or (n > 9) then
           n := ord(toupper(S[idx]))
             - ord('A') + 10;
       if (n < 0) or ( n >= base) then
           break;
       lastResult := result;
       result := (result * base) + n;
       if result < lastResult then
          raise ECstrException.Create
            ('strtol(): overflow!');
       inc(idx);
       end;
    if idx < len then
       endp := @S[idx]
    else
       endp := nil;
    if negate then
       result := - result;
end;

function strtoul(S : PChar; var endp : PChar;
  base : integer) : cardinal;
var x : longint;
begin
    x := strtol(S,endp,base);
    result := cardinal(x);
    if (x < 0) or (result < x) then
       raise ECstrException.Create
         ('strtoul(): overflow!');
end;

function StrPChar(const s: string) : PChar;
begin
  PCharBuf.cpy_string(s);
  result := PCharBuf.Str;
end;

type

     Pcstr = ^cstr;
     cstrTFD = record
          buf : Pcstr;
          idx : cardinal;
          end;
     PcstrTFD = ^cstrTFD;

function cstr_in(var F : TTextRec) : integer; far;
begin
    result := 0;
    with F do begin
        with PcstrTFD(@UserData[1])^ do begin
            BufPos := 0;
            BufEnd := buf^.len - idx;
            if BufEnd > BufSize then
               BufEnd := BufSize;
            if BufEnd > 0 then begin
               move((buf^.str)[idx],BufPtr^,BufEnd);
               inc(idx,BufEnd);
               end
        end;
    end;
end;

function cstr_out(var F : TTextRec) : integer; far;
begin
    result := 0;
    with F do begin
        with PcstrTFD(@UserData[1])^ do begin
            if BufPos > 0 then begin
               buf^.ncat_extra(PChar(BufPtr),BufPos,1);
               buf^.ncat_char_fill(#0,1);
               buf^.lenReCalc;
               BufPos := 0;
               end;
        end;
    end;
end;

function cstr_flushin(var F : TTextRec) : integer; far;
begin
    result := 0;
end;

function cstr_flushout(var F : TTextRec) : integer; far;
begin
    result := cstr_out(F);
end;

function cstr_closein(var F : TTextRec) : integer; far;
begin
    result := 0;
end;

function cstr_closeout(var F : TTextRec) : integer; far;
begin
    result := 0;
end;

function cstr_open(var F : TTextRec) : integer; far;
begin
    result := 0;
    with F do begin
        PcstrTFD(@UserData[1])^.idx := 0;
        case Mode of
          fmInput: begin
            InOutFunc := @cstr_in;
            FlushFunc := @cstr_flushin;
            CloseFunc := @cstr_closein;
            end;
          fmOutput: begin
            InOutFunc := @cstr_out;
            FlushFunc := @cstr_flushout;
            CloseFunc := @cstr_closeout;
            PcstrTFD(@UserData[1])^.buf^.clr;
            end;
          fmInOut: begin
            Mode := fmOutput;
            InOutFunc := @cstr_out;
            FlushFunc := @cstr_flushout;
            CloseFunc := @cstr_closeout;
            end;
        else
            raise ECstrException.Create
              ('cstr_open invalid mode');
        end;
    end;
end;

procedure AssignFile_cstr
  (var F : TextFile; buf: cstr);
begin
  with TTextRec(F) do begin
      Mode := fmClosed;
      BufSize := sizeof(Buffer);
      BufPtr := @Buffer;
      OpenFunc := @cstr_open;
      InOutFunc := nil;
      FlushFunc := nil;
      CloseFunc := nil;
      PcstrTFD(@UserData[1])^.buf := @buf;
      Name[0] := #0;
      end;
end;

destructor CstrStream.Destroy;
begin
  fCstr.Free;
  inherited Destroy;
end;

function CstrStream.Read(var Buffer; Count: longint): longint;
begin
  if Assigned(fCstr) and (count < longint(CSTR_MAX)) then begin
     result := longint(fCstr.readBuf(buffer,count,pos));
     inc(pos,cardinal(result));
     end
  else
     result := 0;
end;

function CstrStream.Write(const Buffer; Count: longint): longint;
begin
  if Assigned(fCstr) and (count < longint(CSTR_MAX)) then begin
     result := longint(fCstr.writeBuf(buffer,count,pos));
     inc(pos,cardinal(result));
     end
  else
     result := 0;
end;

function CstrStream.Seek(Offset : longint; Origin: word): longint;
var len_ : cardinal;
begin
  if Assigned(fCstr) then
     len_ := fCstr.len
  else
     len_ := 0;
  case Origin of
    0: begin
        if OffSet > longint(len_) then
           raise ECstrException.Create('seek out of bounds');
        pos := cardinal(Offset);
       end;
    1: begin
        if ((longint(pos) + offset) < 0) or
           ((longint(pos) + offset) > longint(len_)) then
           raise ECstrException.Create('seek out of bounds');
        Inc(pos, cardinal(Offset));
       end;
    2: begin
        if (Offset > 0) or ((Offset + longint(len_)) < 0) then
           raise ECstrException.Create('seek out of bounds');
        pos := cardinal(longint(len_) + Offset);
       end;
  end;
  Result := longint(pos);
end;

procedure CstrStream.LoadFromStream(Stream: TStream);
begin
  pos := 0;
  if Assigned(Stream) then begin
     if not Assigned(fCstr) then
        fCstr := cstr.Create;
     if Assigned(fCstr) then
        fCstr.readFromStream(Stream);
     end;
end;

procedure CstrStream.LoadFromFile(const FileName: string);
begin
  pos := 0;
  if not Assigned(fCstr) then
     fCstr := cstr.Create;
  if Assigned(fCstr) then
     fCstr.LoadFromFile(filename);
end;

procedure CstrStream.SaveToStream(Stream: TStream);
begin
  if Assigned(Stream) and Assigned(fCstr) then
     fCstr.storeOnStream(Stream);
end;

procedure CstrStream.SaveToFile(const FileName: string);
begin
  if Assigned(fCstr) then
     fCstr.SaveToFile(filename);
end;

procedure CstrStream.Clear;
begin
  if Assigned(fCstr) then
     fCstr.trunc(0);
  pos := 0;
end;

procedure pstrcpy(var s : string; const ct : string);
  begin s := ct; end;
procedure pstrncpy(var s : string; const ct : string; n : integer);
  begin s := Copy(ct,0,n); end;
procedure pstrcat(var s : string; const ct : string);
  begin s := s + ct; end;
procedure pstrncat(var s : string; const ct : string; n : integer);
  begin s := s + Copy(ct,0,n); end;
function  pstrcmp(const cs, ct : string) : integer;
  begin result := AnsiCompareStr(cs,ct); end;
function  pstricmp(const cs, ct : string) : integer;
  begin result := AnsiCompareText(cs,ct); end;
function  pstrncmp(const cs, ct : string; n : integer) : integer;
  begin result := AnsiCompareStr(Copy(cs,0,n),Copy(ct,0,n)); end;
function  pstrncmpi(const cs, ct : string; n : integer) : integer;
  begin result := AnsiCompareText(Copy(cs,0,n),Copy(ct,0,n)); end;
function  pstrchr(const cs : string; c : char) : integer;
var i, len : integer;
begin
  result := 0;
  i := 1;
  len := length(cs);
  while i < len do
    if cs[i] = c then begin
       result := i;
       exit;
       end
    else
      inc(i);
end;
function  pstrrchr(const cs : string; c: char) : integer;
var i : integer;
begin
  result := 0;
  i := length(cs);
  while i > 0 do
    if cs[i] = c then begin
       result := i;
       exit;
       end
    else
       dec(i);
end;
function  pstrspn(const cs, ct : string) : integer;
var i, j, cslen, ctlen : integer;
begin
  result := 0;
  i := 1;
  cslen := length(cs);
  ctlen := length(ct);
  while (i <= cslen) do begin
    j := ctlen;
    while (j > 0) do
      if (cs[i] = ct[j]) then
          break
      else
         dec(j);
    if (cs[i] <> ct[j]) then begin
       result := i - 1;
       exit;
       end;
    inc(i);
    end;
  result := i - 1;
end;
function  pstrcspn(const cs, ct : string) : integer;
var i, j, cslen, ctlen : integer;
begin
  cslen := length(cs);
  result := cslen;
  i := 1;
  ctlen := length(ct);
  while (i <= cslen) do begin
    j := ctlen;
    while (j > 0) do
      if (cs[i] = ct[j]) then begin
         result := i - 1;
         exit;
         end
      else
         dec(j);
    inc(i);
    end;
end;
function  pstrbrk(const cs, ct : string) : integer;
  begin result := pstrcspn(cs,ct) + 1; end;
function  pstrstr(const cs, ct : string) : integer;
var i, cslen, ctlen, last : integer;
begin
  result := 0;
  ctlen := length(ct);
  cslen := length(cs);
  last := cslen - ctlen + 1;
  i := 1;
  while i <= last do
    if AnsiCompareStr(cs[i],ct) = 0 then begin
       result := i;
       exit
       end
    else
       inc(i);
end;
function  pstrlen(const cs : string) : integer;
  begin result := length(cs); end;
procedure ptrimBothWS(var s : string);
var i, len : integer;
begin
  len := length(s);
  if len > 0 then begin
     delete(s,1,pstrspn(s,CSTR_WS)+1);
     i := len;
     while (i > 0) do
       if (pstrcspn(Copy(s,i,len-i+1),CSTR_WS) > 0) then begin
           delete(s,i+1,len-i);
           break;
           end
       else
          dec(i);
     end;
end;

function patof(const s : string) : extended;
var NULL : string;
begin result := pstrtod(s,NULL); end;
function patoi(const s : string) : integer;
var NULL : string;
begin result := integer(pstrtol(s,NULL,10)); end;
function patol(const s : string) : longint;
var NULL : string;
begin result := pstrtol(s,NULL,10); end;
function pstrtod(const s : string; var ends : string) : extended;
var idx, len : integer; N : string;
begin
  len := length(s);
  idx := pstrspn(s,CSTR_WS) + 1;
  if idx > len then
     raise ECstrException.Create
         ('pstrtod(): invalid string');
  if S[idx] = '-' then begin
     N := S[idx];
     inc(idx);
     end;
  while idx <= len do
    if pstrspn(Copy(s,idx,len-idx+1),'.0123456789') = 0 then
       break
    else begin
       N := N + S[idx];
       inc(idx);
       end;
  if idx <= len then
     if S[idx] in ['E','e'] then begin
        N := N + S[idx];
        inc(idx);
        if idx <= len then begin
           if S[idx] in ['-','+'] then
              N := N + S[idx];
              inc(idx);
           end;
           while idx <= len do
             if pstrspn(Copy(s,idx,len-idx+1),'0123456789') = 0 then
                break
             else begin
                N := N + S[idx];
                inc(idx);
                end;
        end;
  result := StrToFloat(N);
  if idx <= len then
     ends := Copy(s,idx,len-idx+1)
  else
     ends := '';
end;
function pstrtol(const s : string; var ends : string;
  base : integer) : longint;
var idx, len, n : integer;
    negate : boolean;
    lastResult : longint;
begin
    result := 0;
    lastResult := result;
    len := length(S);
    idx := pstrspn(S,CSTR_WS) + 1;
    if idx > len then
       raise ECstrException.Create
           ('pstrtol(): invalid string');
    if base = 16 then begin
       if pstrncmpi(Copy(s,idx,len-idx+1),'0X',2) = 0 then
          inc(idx,2);
       end
    else if base = 0 then begin
       if S[idx] = '0' then begin
          inc(idx);
          if idx > len then
             raise ECstrException.Create
               ('pstrtol(): invalid string');
          if S[idx] in ['x','X'] then begin
             inc(idx);
             base := 16;
             end
          else
             base := 8;
          end
       else
          base := 10;
       end;
    if (base < 2) or (base > 36) then
       raise ECstrException.Create
           ('pstrtol() : invalid base');
    if idx > len then
       raise ECstrException.Create
           ('pstrtol(): invalid string');
    if S[idx] = '-' then begin
       negate := true;
       inc(idx);
       if idx > len then
          raise ECstrException.Create
              ('pstrtol(): invalid string');
       end
    else
       negate := false;
    while idx <= len do begin
       n := ord(S[idx]) - ord('0');
       if (n < 0) or (n > 9) then
           n := ord(toupper(S[idx]))
             - ord('A') + 10;
       if (n < 0) or ( n >= base) then
           break;
       lastResult := result;
       result := (result * base) + n;
       if result < lastResult then
          raise ECstrException.Create
            ('pstrtol(): overflow!');
       inc(idx);
       end;
    if idx <= len then
       ends := Copy(s,idx,len-idx+1)
    else
       ends := '';
    if negate then
       result := - result;
end;
function pstrtoul(const s : string; var ends : string;
  base : integer) : cardinal;
var x : longint;
begin
    x := pstrtol(S,ends,base);
    result := cardinal(x);
    if (x < 0) or (result < x) then
       raise ECstrException.Create
         ('pstrtoul(): overflow!');
end;

var ExitSave : Pointer;

procedure UnitInit;
begin
    strtok_input := cstr.Create;
    strtok_found := cstr.Create;
    fill_char := cstr.Create;
    PCharBuf := cstr.Create;
end;

procedure UnitDone;
begin
    strtok_input.Free;
    strtok_found.Free;
    fill_char.Free;
    PCharBuf.Free;
end;

procedure UnitExit; far;
begin
    ExitProc := ExitSave;
    UnitDone;
end;

begin
    ExitSave := ExitProc;
    ExitProc := @UnitExit;
    UnitInit;
end.

